@@ -1,4 +1,4 @@
-from itertools import repeat
+from itertools import repeat, chain
 from typing import List, Optional, Dict
 
 from conditional import conditional
@@ -14,7 +14,7 @@ from model_clone.utils import (
     transaction_autocommit,
     get_unique_value,
     context_mutable_attribute,
-    unpack_unique_together,
+    get_fields_and_unique_fields_from_cls,
 )
 
 
@@ -24,23 +24,23 @@ class CloneMixin(object):
     CloneMixin mixin to duplicate an object using the model cls.
 
     :param _clone_fields: Restricted List of fields to copy from the instance.
-    :type _clone_fields`collections.Iterable`
+    :type _clone_fields: collections.Iterable
     :param _clone_m2m_fields: Many to many fields (Example: TestModel.tags).
-    :type _clone_m2m_fields`collections.Iterable`
+    :type _clone_m2m_fields: collections.Iterable
     :param _clone_m2o_or_o2m_fields: Many to one/One to many fields.
-    :type _clone_m2o_or_o2m_fields`collections.Iterable`
+    :type _clone_m2o_or_o2m_fields: collections.Iterable
     :param _clone_o2o_fields: One to One fields.
-    :type _clone_o2o_fields`collections.Iterable`
+    :type _clone_o2o_fields: collections.Iterable
 
     :param _clone_excluded_fields: Excluded model fields.
-    :type _clone_excluded_fields`collections.Iterable`
+    :type _clone_excluded_fields: collections.Iterable
     :param _clone_excluded_m2m_fields: Excluded many to many fields.
-    :type _clone_excluded_m2m_fields`collections.Iterable`
+    :type _clone_excluded_m2m_fields: collections.Iterable
     :param _clone_excluded_m2o_or_o2m_fields: Excluded many to many
         and one to many fields.
-    :type _clone_excluded_m2o_or_o2m_fields`collections.Iterable`
+    :type _clone_excluded_m2o_or_o2m_fields: collections.Iterable
     :param _clone_excluded_o2o_fields: Excluded one to one fields.
-    :type _clone_excluded_o2o_fields: `collections.Iterable`
+    :type _clone_excluded_o2o_fields: collections.Iterable
 
     :Example:
     >>> # Using explicit fields
@@ -94,7 +94,7 @@ class CloneMixin(object):
     MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS = 100  # type: int
 
     @staticmethod
-    def _create_copy_of_instance(instance, force=False):
+    def _create_copy_of_instance(instance, force=False, sub_clone=False):
         """
         Create a copy of an instance
 
@@ -102,6 +102,8 @@ class CloneMixin(object):
         :type instance: `django.db.models.Model`
         :param force: Flag to skip using the current model clone declared attributes.
         :type force: bool
+        :param sub_clone: Flag to skip cloning one to one fields for sub clones.
+        :type sub_clone: bool
         :return: A new transient instance.
         :rtype: `django.db.models.Model`
         """
@@ -110,6 +112,12 @@ class CloneMixin(object):
         clone_excluded_fields = getattr(
             cls, "_clone_excluded_fields", CloneMixin._clone_excluded_fields
         )
+        clone_o2o_fields = getattr(
+            cls, "_clone_o2o_fields", CloneMixin._clone_o2o_fields
+        )
+        clone_excluded_o2o_fields = getattr(
+            cls, "_clone_excluded_o2o_fields", CloneMixin._clone_excluded_o2o_fields
+        )
         unique_duplicate_suffix = getattr(
             cls, "UNIQUE_DUPLICATE_SUFFIX", CloneMixin.UNIQUE_DUPLICATE_SUFFIX
         )
@@ -124,32 +132,15 @@ class CloneMixin(object):
             CloneMixin.MAX_UNIQUE_DUPLICATE_QUERY_ATTEMPTS,
         )
 
-        fields = []
-
-        for f in instance._meta.concrete_fields:
-            valid = False
-            if not f.primary_key:
-                if clone_fields and not force:
-                    valid = f.name in clone_fields
-                elif clone_excluded_fields and not force:
-                    valid = f.name not in clone_excluded_fields
-                else:
-                    valid = True
-
-            if valid:
-                fields.append(f)
-
-        unique_field_names = unpack_unique_together(
-            opts=instance._meta,
-            only_fields=[f.attname for f in fields],
+        fields, unique_fields = get_fields_and_unique_fields_from_cls(
+            cls=cls,
+            force=force,
+            clone_fields=clone_fields,
+            clone_excluded_fields=clone_excluded_fields,
+            clone_o2o_fields=clone_o2o_fields,
+            clone_excluded_o2o_fields=clone_excluded_o2o_fields,
         )
 
-        unique_fields = [
-            f.name
-            for f in fields
-            if not f.auto_created and (f.unique or f.name in unique_field_names)
-        ]
-
         new_instance = cls()
 
         for f in fields:
@@ -165,14 +156,9 @@ class CloneMixin(object):
                     not f.auto_created,
                     f.concrete,
                     f.editable,
-                    not isinstance(f, (models.DateTimeField, models.DateField)),
-                    f not in instance._meta.related_objects,
-                    f not in instance._meta.many_to_many,
                     f.name in unique_fields,
                 ]
             ):
-                value = getattr(instance, f.attname, f.get_default())
-
                 # Do not try to get unique value for enum type field
                 if isinstance(f, models.CharField) and not f.choices:
                     value = clean_value(value, unique_duplicate_suffix)
@@ -187,6 +173,18 @@ class CloneMixin(object):
                             max_attempts=max_unique_duplicate_query_attempts,
                         )
 
+                elif isinstance(f, models.OneToOneField) and not sub_clone:
+                    sub_instance = getattr(instance, f.name, f.get_default())
+
+                    if sub_instance is not None:
+                        sub_instance = CloneMixin._create_copy_of_instance(
+                            sub_instance,
+                            force=True,
+                            sub_clone=True,
+                        )
+                        sub_instance.save()
+                        value = sub_instance.pk
+
             setattr(new_instance, f.attname, value)
 
         return new_instance
@@ -277,126 +275,63 @@ class CloneMixin(object):
 
         return errors
 
-    @transaction.atomic
-    def make_clone(self, attrs=None, sub_clone=False):
+    def __duplicate_o2o_fields(self, duplicate):
         """
-        Creates a clone of the django model instance.
+        Duplicate the one to one fields.
 
-        :param attrs (dict): Dictionary of attributes to be replaced on the cloned object.
-        :param sub_clone (bool): Internal boolean used to detect cloning sub objects.
-        :rtype: :obj:`django.db.models.Model`
-        :return: The model instance that has been cloned.
+        :param duplicate: The transient instance that should be duplicated.
+        :type duplicate: `django.db.models.Model`
+        :return: The duplicate instance with all the one to one fields duplicated.
         """
-        attrs = attrs or {}
-        if not self.pk:
-            raise ValidationError(
-                "{}: Instance must be saved before it can be cloned.".format(
-                    self.__class__.__name__
-                )
-            )
-        if sub_clone:
-            duplicate = self
-            duplicate.pk = None
-        else:
-            duplicate = self._create_copy_of_instance(self)
-            # Supports only updating the attributes of the base instance.
-            for name, value in attrs.items():
-                setattr(duplicate, name, value)
-
-        one_to_one_fields = [
-            f for f in self._meta.concrete_fields if getattr(f, "one_to_one", False)
-        ]
-        many_to_one_or_one_to_many_fields = []
-        many_to_many_fields = []
-
-        for f in self._meta.related_objects:
-            if f.one_to_one and f.name in self._clone_o2o_fields:
-                one_to_one_fields.append(f)
-
-            elif all(
-                [
-                    not self._clone_o2o_fields,
-                    f.one_to_one,
-                    self._clone_excluded_o2o_fields,
-                    f not in one_to_one_fields,
-                    f.name not in self._clone_excluded_o2o_fields,
-                ]
-            ):
-                one_to_one_fields.append(f)
-
-            elif all(
-                [
-                    any([f.many_to_one, f.one_to_many]),
-                    f.name in self._clone_m2o_or_o2m_fields,
-                ]
-            ):
-                many_to_one_or_one_to_many_fields.append(f)
-
-            elif all(
-                [
-                    not self._clone_m2o_or_o2m_fields,
-                    any([f.many_to_one, f.one_to_many]),
-                    self._clone_excluded_m2o_or_o2m_fields,
-                    f not in many_to_one_or_one_to_many_fields,
-                    f.name not in self._clone_excluded_m2o_or_o2m_fields,
-                ]
-            ):
-                many_to_one_or_one_to_many_fields.append(f)
-
-            elif all(
-                [
-                    f.many_to_many,
-                    f.name in self._clone_m2m_fields,
-                ]
-            ):
-                many_to_many_fields.append(f)
-
-            elif all(
-                [
-                    f.many_to_many,
-                    not self._clone_m2m_fields,
-                    self._clone_excluded_m2m_fields,
-                    f not in many_to_many_fields,
-                    f.name not in self._clone_excluded_m2m_fields,
-                ]
-            ):
-                many_to_many_fields.append(f)
-
-        for f in self._meta.many_to_many:
-            if not sub_clone:
-                if f.name in self._clone_m2m_fields:
-                    many_to_many_fields.append(f)
-                elif all(
+        if self._clone_o2o_fields or self._clone_excluded_o2o_fields:
+            for f in self._meta.related_objects:
+                if any(
                     [
-                        not self._clone_m2m_fields,
-                        self._clone_excluded_m2m_fields,
-                        f.name not in self._clone_excluded_m2m_fields,
-                        f not in many_to_many_fields,
+                        f.one_to_one
+                        and f.name in self._clone_o2o_fields
+                        and f not in self._meta.concrete_fields,
+                        f.one_to_one
+                        and self._clone_excluded_o2o_fields
+                        and f.name not in self._clone_excluded_o2o_fields
+                        and f not in self._meta.concrete_fields,
                     ]
                 ):
-                    many_to_many_fields.append(f)
+                    rel_object = getattr(self, f.name, None)
+                    if rel_object:
+                        new_rel_object = CloneMixin._create_copy_of_instance(
+                            rel_object,
+                            force=True,
+                            sub_clone=True,
+                        )
+                        setattr(new_rel_object, f.remote_field.name, duplicate)
+                        new_rel_object.save()
 
-        # Clone one to one fields
-        for field in one_to_one_fields:
-            rel_object = getattr(self, field.name, None)
-            new_rel_object = None
-            if rel_object:
-                if hasattr(rel_object, "make_clone") and callable(
-                    rel_object.make_clone
+        return duplicate
+
+    def __duplicate_o2m_m2o_fields(self, duplicate):
+        """
+        Duplicate many to one or one to many fields.
+
+        :param duplicate: The transient instance that should be duplicated.
+        :type duplicate: `django.db.models.Model`
+        :return: The duplicate instance with all the many to one or one to many fields duplicated.
+        """
+        fields = set()
+
+        if self._clone_m2o_or_o2m_fields or self._clone_excluded_m2o_or_o2m_fields:
+            for f in self._meta.related_objects:
+                if any(
+                    [
+                        any([f.many_to_one, f.one_to_many])
+                        and f.name in self._clone_m2o_or_o2m_fields,
+                        any([f.many_to_one, f.one_to_many])
+                        and f.name not in self._clone_excluded_m2o_or_o2m_fields,
+                    ]
                 ):
-                    new_rel_object = rel_object.make_clone(sub_clone=True)
-                else:
-                    new_rel_object = CloneMixin._create_copy_of_instance(
-                        rel_object, force=True
-                    )
-                    new_rel_object.save()
-
-            setattr(duplicate, field.name, new_rel_object)
-
-        duplicate.save()
+                    fields.add(f)
 
         # Clone one to many/many to one fields
-        for field in many_to_one_or_one_to_many_fields:
+        for field in fields:
             items = []
             for item in getattr(self, field.related_name).all():
                 try:
@@ -411,8 +346,37 @@ class CloneMixin(object):
 
             getattr(duplicate, field.related_name).set(items)
 
+        return duplicate
+
+    def __duplicate_m2m_fields(self, duplicate, sub_clone):
+        """
+        Duplicate many to many fields.
+
+        :param duplicate: The transient instance that should be duplicated.
+        :type duplicate: `django.db.models.Model`
+        :param sub_clone: Boolean indicating that the instance is a sub instance
+            i.e a related field of a top level instance.
+        :return: The duplicate instance with all the many to many fields duplicated.
+        """
+        fields = set()
+
+        # Duplicating sub instance many to many fields not currently supported.
+        if sub_clone:
+            return duplicate
+
+        if self._clone_m2m_fields or self._clone_excluded_m2m_fields:
+            for f in chain(self._meta.related_objects, self._meta.many_to_many):
+                if any(
+                    [
+                        f.many_to_many and f.name in self._clone_m2m_fields,
+                        f.many_to_many
+                        and f.name not in self._clone_excluded_m2m_fields,
+                    ]
+                ):
+                    fields.add(f)
+
         # Clone many to many fields
-        for field in many_to_many_fields:
+        for field in fields:
             if hasattr(field, "field"):
                 # ManyToManyRel
                 field_name = field.field.m2m_reverse_field_name()
@@ -440,6 +404,43 @@ class CloneMixin(object):
                         item.save()
             else:
                 destination.set(source.all())
+
+        return duplicate
+
+    @transaction.atomic
+    def make_clone(self, attrs=None, sub_clone=False):
+        """
+        Creates a clone of the django model instance.
+
+        :param attrs: Dictionary of attributes to be replaced on the cloned object.
+        :type attrs: dict
+        :param sub_clone: Internal boolean used to detect cloning sub objects.
+        :type sub_clone: bool
+        :rtype: :obj:`django.db.models.Model`
+        :return: The model instance that has been cloned.
+        """
+        attrs = attrs or {}
+        if not self.pk:
+            raise ValidationError(
+                "{}: Instance must be saved before it can be cloned.".format(
+                    self.__class__.__name__
+                )
+            )
+        if sub_clone:
+            duplicate = self
+            duplicate.pk = None
+        else:
+            duplicate = self._create_copy_of_instance(self)
+            # Supports only updating the attributes of the base instance.
+            for name, value in attrs.items():
+                setattr(duplicate, name, value)
+
+        duplicate.full_clean()
+        duplicate.save()
+
+        duplicate = self.__duplicate_o2o_fields(duplicate)
+        duplicate = self.__duplicate_o2m_m2o_fields(duplicate)
+        duplicate = self.__duplicate_m2m_fields(duplicate, sub_clone)
         return duplicate
 
     @classmethod
